/*
 ftd2.h - Serial Library for FTDI USB Chip FT2XX

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "ftd2xx.h"
#include "sleep.h"

#ifdef __linux__
  #define DEVICE_NAME "ttyUSB"
  #define prefixcmp strncmp
#else
  #define DEVICE_NAME "COM"
  #define prefixcmp strnicmp
#endif

#ifndef LOW
#define LOW  0
#endif
#ifndef HIGH
#define HIGH 1
#endif

extern FT_STATUS FT_GetLastError(void);

class FTD2
{
  public:

    //
    // FT_OpenEx Flags
    //
    typedef enum
    {
      OPENBY_SERIAL_NUMBER = FT_OPEN_BY_SERIAL_NUMBER,
      OPENBY_DESCRIPTION   = FT_OPEN_BY_DESCRIPTION,
      OPENBY_LOCATION      = FT_OPEN_BY_LOCATION,
    } OPENBY;
    //
    // Baud Rates
    //
    typedef enum
    {
      SPEED_300     = FT_BAUD_300,
      SPEED_600     = FT_BAUD_600,
      SPEED_1200    = FT_BAUD_1200,
      SPEED_2400    = FT_BAUD_2400,
      SPEED_4800    = FT_BAUD_4800,
      SPEED_9600    = FT_BAUD_9600,
      SPEED_14400   = FT_BAUD_14400,
      SPEED_19200   = FT_BAUD_19200,
      SPEED_38400   = FT_BAUD_38400,
      SPEED_57600   = FT_BAUD_57600,
      SPEED_115200  = FT_BAUD_115200,
      SPEED_230400  = FT_BAUD_230400,
      SPEED_460800  = FT_BAUD_460800,
      SPEED_500K    = 500000,
      SPEED_921600  = FT_BAUD_921600,
      SPEED_1M      = 1000000,
      SPEED_2M      = 2000000,
      SPEED_3M      = 3000000,
    } SPEED;
    //
    // mode
    //
    typedef enum
    {
      MODE_7N1 = 0,
      MODE_8N1 = 1,
      MODE_7N2 = 2,
      MODE_8N2 = 3,
      MODE_7E1 = 4,
      MODE_8E1 = 5,
      MODE_7E2 = 6,
      MODE_8E2 = 7,
      MODE_7O1 = 8,
      MODE_8O1 = 9,
      MODE_7O2 = 10,
      MODE_8O2 = 11,
    } MODE;
    //
    // Flow Control
    //
    typedef enum
    {
      FLOW_NONE     = FT_FLOW_NONE,
      FLOW_RTS_CTS  = FT_FLOW_RTS_CTS,
      FLOW_DTR_DSR  = FT_FLOW_DTR_DSR,
      FLOW_XON_XOFF = FT_FLOW_XON_XOFF,
    } FLOW;
    //
    // Bit Modes
    //
    typedef enum
    {
      BITMODE_RESET         = FT_BITMODE_RESET,
      BITMODE_ASYNC_BITBANG = FT_BITMODE_ASYNC_BITBANG,
      BITMODE_MPSSE         = FT_BITMODE_MPSSE,
      BITMODE_SYNC_BITBANG  = FT_BITMODE_SYNC_BITBANG,
      BITMODE_MCU_HOST      = FT_BITMODE_MCU_HOST,
      BITMODE_FAST_SERIAL   = FT_BITMODE_FAST_SERIAL,
      BITMODE_CBUS_BITBANG  = FT_BITMODE_CBUS_BITBANG,
      BITMODE_SYNC_FIFO     = FT_BITMODE_SYNC_FIFO,
    } BITMODE;

    typedef enum
    {
      GPIO_INPUT  = 0,
      GPIO_OUTPUT = 1,
    } GPIO;

    typedef enum
    {
      CBUS_0 = 0,
      CBUS_1 = 1,
      CBUS_2 = 2,
      CBUS_3 = 3,
    } CBUS;

    FTD2(BITMODE mode = BITMODE_CBUS_BITBANG)
    : _handle(0)
    , _mode(mode)
    , _cbus(0)
    {
      FT_Initialise();
    }

    virtual ~FTD2(void)
    {
      close();
      FT_Finalise();
    }

    FT_STATUS status(void)
    {
      return FT_GetLastError();
    }

    int getDeviceCount(void)
    {
      int count = 0;
      FT_ListDevices(&count, NULL, FT_LIST_NUMBER_ONLY);
      return count;
    }

    int getComPortNumber(int index)
    {
      long num = -1;
      FT_HANDLE handle;
      if (FT_SUCCESS(FT_Open(index, &handle)))
      {
        FT_GetComPortNumber(handle, &num);
        FT_Close(handle);
      }
      return num;
    }

    bool getComPortName(int index, char *buf)
    {
      int num;
      strcpy(buf, "(null)");
      if ((num = getComPortNumber(index)) >= 0)
      {
        sprintf(buf, DEVICE_NAME "%d", num);
        return true;
      }
      return false;
    }

    bool getDeviceSerial(int index, char *buf)
    {
      strcpy(buf, "(null)");
      return FT_SUCCESS(FT_ListDevices((void *)(size_t)index, buf, FT_LIST_BY_INDEX | FT_OPEN_BY_SERIAL_NUMBER));
    }

    bool getDeviceDescription(int index, char *buf)
    {
      strcpy(buf, "(null)");
      return FT_SUCCESS(FT_ListDevices((void *)(size_t)index, buf, FT_LIST_BY_INDEX | FT_OPEN_BY_DESCRIPTION));
    }

    bool getDeviceLocation(int index, char *buf)
    {
      strcpy(buf, "(null)");
      return FT_SUCCESS(FT_ListDevices((void *)(size_t)index, buf, FT_LIST_BY_INDEX | FT_OPEN_BY_LOCATION));
    }

    bool open(const char *device, SPEED speed = SPEED_115200, MODE mode = MODE_8N1, FLOW flow = FLOW_NONE, uint32_t timeout = 1000)
    {
      static const char PREFIX[] = DEVICE_NAME;
      char *eos;
      int index = strtoul(device, &eos, 10);
      if (*eos)
      {
        if (prefixcmp(device, PREFIX, sizeof(PREFIX) - 1) == 0)
        {
          int num = strtoul(device + sizeof(PREFIX) - 1, &eos, 10);
          if (*eos == 0)
          {
            int count = getDeviceCount();
            for (index = 0; index < count; ++index)
            {
              if (getComPortNumber(index) == num)
              {
                if (FT_SUCCESS(FT_Open(index, &_handle)))
                {
                  init(speed, mode, flow, timeout);
                  return true;
                }
              }
            }
          }
        }
        if (!FT_SUCCESS(FT_OpenEx((void *)device, OPENBY_SERIAL_NUMBER, &_handle)))
          return false;
      }
      else
      {
        if (!FT_SUCCESS(FT_Open(index, &_handle)))
          return false;
      }
      init(speed, mode, flow, timeout);
      return true;
    }

    bool close(void)
    {
      if (_handle)
      {
        FT_HANDLE h = _handle;
        _handle = nullptr;
        return FT_SUCCESS(FT_Close(h));
      }
      return true;
    }

    int getComPortNumber(void)
    {
      LONG com = -1;
      FT_GetComPortNumber(_handle, &com);
      return com;
    }

    bool baudrate(SPEED speed)
    {
      switch (speed)
      {
        case SPEED_500K: return FT_SUCCESS(FT_SetDivisor(_handle, 6));
        case SPEED_1M:   return FT_SUCCESS(FT_SetDivisor(_handle, 3));
        case SPEED_2M:   return FT_SUCCESS(FT_SetDivisor(_handle, 1));
        case SPEED_3M:   return FT_SUCCESS(FT_SetDivisor(_handle, 0));
        default: break;
      }
      return FT_SUCCESS(FT_SetBaudRate(_handle, speed));
    }

    bool frame(MODE mode = MODE_8N1)
    {
      return FT_SUCCESS(FT_SetDataCharacteristics(_handle,
        mode & 1 ? DATA_8   : DATA_7,
        mode & 2 ? STOP_2   : STOP_1,
        mode & 4 ? PAR_EVEN : (mode & 8 ? PAR_NONE : PAR_NONE)
      ));
    }

    bool flowCtrl(FLOW mode, uint8_t xon = 0x11, uint8_t xoff = 0x13)
    {
      return FT_SUCCESS(FT_SetFlowControl(_handle, mode, xon, xoff));
    }

    bool dtr(bool on)
    {
      return FT_SUCCESS(on ? FT_SetDtr(_handle) : FT_ClrDtr(_handle));
    }

    bool rts(bool on)
    {
      return FT_SUCCESS(on ? FT_SetRts(_handle) : FT_ClrRts(_handle));
    }

    bool setTimeout(uint32_t readTimeout, uint32_t writeTimeout)
    {
      return FT_SUCCESS(FT_SetTimeouts(_handle, readTimeout, writeTimeout));
    }

    bool purge(void)
    {
      return FT_SUCCESS(FT_Purge(_handle, FT_PURGE_RX | FT_PURGE_TX));
    }

    size_t read(void *buf, size_t len)
    {
      DWORD bytesReceived = 0;
      FT_Read(_handle, buf, len, &bytesReceived);
      return bytesReceived;
    }

    size_t write(const void *buf, size_t len)
    {
      DWORD bytesWritten = 0;
      FT_Write(_handle, (void *)buf, len, &bytesWritten);
      return bytesWritten;
    }

    void pinMode(CBUS pin, GPIO mode)
    {
      pin = (CBUS)((pin & 3) + 4);
      if (mode == GPIO_OUTPUT)
        _cbus |=  (1 << pin);
      else
        _cbus &= ~(1 << pin);
      FT_SetBitMode(_handle, _cbus, _mode);
    }

    void digitalWrite(CBUS pin, uint8_t value)
    {
      pin = (CBUS)(pin & 3);
      if (value)
        _cbus |=  (1 << pin);
      else
        _cbus &= ~(1 << pin);
      FT_SetBitMode(_handle, _cbus, _mode);
    }

    uint8_t digitalRead(CBUS pin)
    {
      pin = (CBUS)(pin & 3);
      uint8_t mask = _cbus >> 4;
      uint8_t data = 0;
      FT_GetBitMode(_handle, &data);
      return (((data & ~mask) | (_cbus & mask)) >> pin) & 1;
    }

  protected:

    void init(SPEED speed, MODE mode, FLOW flow, uint32_t timeout)
    {
      baudrate(speed);
      frame(mode);
      //flowCtrl(flow);
      rts(true);
      dtr(true);
      setTimeout(timeout, timeout);
      purge();
      FT_SetBitMode(_handle, _cbus, BITMODE_RESET);
    }

  private:

    //
    // Word Lengths
    //
    typedef enum
    {
      DATA_8 = FT_BITS_8,
      DATA_7 = FT_BITS_7,
    } DATA;
    //
    // Stop Bits
    //
    typedef enum
    {
      STOP_1 = FT_STOP_BITS_1,
      STOP_2 = FT_STOP_BITS_2,
    } STOP;
    //
    // Parity
    //
    typedef enum
    {
      PAR_NONE  = FT_PARITY_NONE,
      PAR_ODD   = FT_PARITY_ODD,
      PAR_EVEN  = FT_PARITY_EVEN,
      PAR_MARK  = FT_PARITY_MARK,
      PAR_SPACE = FT_PARITY_SPACE,
    } PAR;

    FT_HANDLE _handle;
    uint8_t   _mode;
    uint8_t   _cbus;
};
