/*
  jn516x.h - JN516x Chip Library for NXP JN516x Series

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stdint.h>

#define FLASH_START_ADDRESS      0x80000U
#define FLASH_SECTOR_OFFSET(sec) (0x8000U * (sec))

#if defined(JN5161)
  #define CHIP_VERSION JN516x::VERSION_JN5161
#elif defined(JN5164)
  #define CHIP_VERSION JN516x::VERSION_JN5164
#elif defined(JN5168)
  #define CHIP_VERSION JN516x::VERSION_JN5168
#elif defined(JN5169)
  #define CHIP_VERSION JN516x::VERSION_JN5169
#elif defined(JN5179)
  #define CHIP_VERSION JN516x::VERSION_JN5179
#else
  #define CHIP_VERSION 0
#endif

class JN516x
{
  public:

    static const uint8_t MAGICNUMBER[12];
    static const uint8_t MAGICNUMBER_JN5179[12];

    typedef struct __attribute__((packed))
    {
      uint8_t  magicNumber[12];
      uint8_t  configuration;
      uint8_t  status;
      uint16_t applicationID;
    } tsBIR;

    typedef struct __attribute__((packed))
    {
      tsBIR    sBIR;           // 16-byte Boot Image Record
      uint8_t  encryption[14]; // Encryption Initialisation Vector (ignored if unencrypted)
      uint16_t configuration;  // 16-bit Software Configuration Options
      uint32_t imageSize;      // 32-bit Length of Binary Image in bytes
      uint32_t dataStart;      // 32-bit .data section Flash start address
      uint16_t dataAddress;    // 16-bit .data section load address in RAM (word aligned)
      uint16_t dataLength;     // 16-bit .data section length in 32-bit words
      uint16_t bssStart;       // 16 bit .bss section start address in RAM (word aligned)
      uint16_t bssLength;      // 16-bit .bss section length in 32-bit words
      uint32_t warmStart;      // 32-bit wake-up entry point (word aligned) – warm start
      uint32_t coldStart;      // 32-bit reset entry point (word aligned) – cold start
    } tsFlashHeader;

    typedef struct __attribute__((packed))
    {
      uint32_t version;        // version
      tsBIR    sBIR;           // 16-byte Boot Image Record
      uint8_t  encryption[14]; // Encryption Initialisation Vector (ignored if unencrypted)
      uint16_t configuration;  // 16-bit Software Configuration Options
      uint32_t imageSize;      // 32-bit Length of Binary Image in bytes
      uint32_t dataStart;      // 32-bit .data section Flash start address
      uint16_t dataAddress;    // 16-bit .data section load address in RAM (word aligned)
      uint16_t dataLength;     // 16-bit .data section length in 32-bit words
      uint16_t bssStart;       // 16 bit .bss section start address in RAM (word aligned)
      uint16_t bssLength;      // 16-bit .bss section length in 32-bit words
      uint32_t warmStart;      // 32-bit wake-up entry point (word aligned) – warm start
      uint32_t coldStart;      // 32-bit reset entry point (word aligned) – cold start
    } tsFlashHeader_JN5179;

    typedef enum
    {
      CHIPID_JN5179 = 1U,          // Survey required.
      CHIPID_JN5169 = 0x0100B686U,
      CHIPID_JN5168 = 0x10008686U,
      CHIPID_JN5164 = 0x10408686U,
      CHIPID_JN5161 = 0U,          // Survey required.
      CHIPID_JN5148 = 0x10404686U,
      CHIPID_JN5142 = 0x00005686U,
      CHIPID_JN5139 = 0x10002000U,
    } CHIPID;

    typedef enum
    {
      VERSION_JN5179 = 0x0F03000AU,
      VERSION_JN5169 = 0x0F03000BU,
      VERSION_JN5168 = 0x07030008U,
      VERSION_JN5164 = 0x04030008U,
      VERSION_JN5161 = 0x01000008U,
    } VERSION;

    typedef enum
    {
      FLASHID_ST_M25P05_A    = 0x0505U,
      FLASHID_ST_M25P10_A    = 0x1010U,
      FLASHID_ST_M25P20_A    = 0x1101U,
      FLASHID_ST_M25P40      = 0x1212U,
      FLASHID_SST_25VF010A   = 0xBF49U,
      FLASHID_ATMEL_25F512   = 0x1F60U,
      FLASHID_INTERNAL_FLASH = 0xCCEEU, // JN516x/7x only
    } FLASHID;

    typedef enum
    {
      FLASH_ST_M25P05_A    = 4U,
      FLASH_ST_M25P10_A    = 0U,
      FLASH_ST_M25P20_A    = 5U,
      FLASH_ST_M25P40      = 3U,
      FLASH_SST_25VF010A   = 1U,
      FLASH_ATMEL_25F512   = 2U,
      FLASH_INTERNAL_FLASH = 8U, // JN516x/7x only
    } FLASH;

    typedef enum
    {
      ADDR_BROADCAST   = 0xFFFFU,
      ADDR_OTA_BOOT    = ADDR_BROADCAST - 1,
      ADDR_OTA_BRIDGE  = ADDR_BROADCAST - 2,
      ADDR_OTA_BRIDGE2 = ADDR_BROADCAST - 3,
    } ADDR;

    typedef enum
    {
      OTA_UPDATE_CHIPVER  = 0xC0U,
      OTA_UPDATE_BLDMODE  = 0xC1U,
      OTA_UPDATE_START    = 0xC2U,
      OTA_UPDATE_PROGRAM  = 0xC3U,
      OTA_UPDATE_END      = 0xC4U,
    } OTA_UPDATE;

    typedef enum
    {
      OTA_STATUS_OK        = 0x00U,
      OTA_STATUS_ARG_ERROR = 0xFAU,
      OTA_STATUS_BLD_ERROR = 0xFBU,
      OTA_STATUS_CRC_ERROR = 0xFCU,
      OTA_STATUS_SDK_ERROR = 0xFDU,
    } OTA_STATUS;

    static
    VERSION chipVersion(CHIPID id)
    {
      switch (id)
      {
        case CHIPID_JN5179: return VERSION_JN5179;
        case CHIPID_JN5169: return VERSION_JN5169;
        case CHIPID_JN5168: return VERSION_JN5168;
        case CHIPID_JN5164: return VERSION_JN5164;
        case CHIPID_JN5161: return VERSION_JN5161;
        default: break;
      }
      return (VERSION)0;
    }

    static
    const char *chipName(VERSION version)
    {
      switch (version)
      {
        case VERSION_JN5179: return "JN5179";
        case VERSION_JN5169: return "JN5169";
        case VERSION_JN5168: return "JN5168";
        case VERSION_JN5164: return "JN5164";
        case VERSION_JN5161: return "JN5161";
        default: break;
      }
      return "(unknown)";
    }

    static
    uint32_t chipFlashSectors(VERSION version)
    {
      return (uint8_t)(version >> 24) + 1;
    }

    static
    uint32_t chipFlashSize(VERSION version)
    {
      return 32768U * chipFlashSectors(version);
    }

    static
    uint32_t chipRamSize(VERSION version)
    {
      return 8192U * ((uint8_t)(version >> 16) + 1);
    }

    static
    uint16_t chipType(VERSION version)
    {
      return (uint16_t)version;
    }
};
