/*
 jn516x_ota.h - OTA Library for NXP JN516x Series

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#ifdef __linux__
#include <stddef.h>
#include "ftdisio.h"
#else
#include "ftd2.h"
#endif
#include "crc32.h"
#include "sleep.h"
#include "jn516x.h"
#include "bridge.h"

#define DEBUG_TX_PACKET_DUMP 0
#define DEBUG_RX_PACKET_DUMP 0
#define OTA_RETRY_COUNT      10

#define OTA_BOOT_FIRMWARE "-OTA-BOOT.bin"

class JN516x_OTA : public JN516x
#ifdef __linux__
, public FTDISIO
#else
, public FTD2
#endif
{
  public:

    typedef enum
    {
      STATUS_OK             = OTA_STATUS_OK,
      STATUS_ARG_ERROR      = OTA_STATUS_ARG_ERROR,
      STATUS_BLD_ERROR      = OTA_STATUS_BLD_ERROR,
      STATUS_CRC_ERROR      = OTA_STATUS_CRC_ERROR,
      STATUS_SDK_ERROR      = OTA_STATUS_SDK_ERROR,
      STATUS_INVALID_DEVICE = 0xF0,
      STATUS_INVALID_FILE   = 0xF1,
      STATUS_OPEN_ERROR     = 0xF2,
      STATUS_READ_ERROR     = 0xF3,
      STATUS_TX_ERROR       = 0xF4,
      STATUS_RX_ERROR       = 0xF5,
    } STATUS;

    JN516x_OTA(void)
    : _address(0)
    , _otaboot(false)
    {
    }

    virtual ~JN516x_OTA(void)
    {
    }

    bool open(const char *portname, unsigned long speed, uint16_t addr)
    {
      _address = addr;
#ifdef __linux
      return FTDISIO::open(portname, speed, SERIAL_FRAME_8N1, SERIAL_FLOW_NONE, 1000) >= 0;
#else
      return FTD2::open(portname, (FTD2::SPEED)speed, MODE_8N1, FLOW_NONE, 1000);
#endif
    }

    STATUS update(const char *filename)
    {
      STATUS status;
      VERSION chipver;
      status = chipVersion(chipver);
      programStart(chipName(chipver), filename, true);
      if (status == STATUS_OK)
      {
        if (!chipver)
          status = STATUS_INVALID_DEVICE;
        else
        {
          FILE *fp;
          status = STATUS_OPEN_ERROR;
          if ((fp = fopen(filename, "rb")))
          {
            VERSION filever;
            if (!fileVersion(filever, fp))
              status = STATUS_READ_ERROR;
            else if (chipType(chipver) != chipType(filever))
              status = STATUS_INVALID_FILE;
            else
            {
              size_t fpos = (chipver == VERSION_JN5179 ? 0 : ftell(fp));
              fseek(fp, 0, SEEK_END);
              size_t size = ftell(fp) - fpos;
              fseek(fp, fpos, SEEK_SET);
              if (size > FLASH_SECTOR_OFFSET(chipFlashSectors(chipver) - 1))
                status = STATUS_INVALID_FILE;
              else if (!strstr(filename, OTA_BOOT_FIRMWARE))
                status = update(fp, size, _address);        // Update App
              else if (size <= FLASH_SECTOR_OFFSET(1))
                status = update(fp, size, _address, true);  // Update OTA-BOOT
              else
                status = STATUS_INVALID_FILE;
            }
            fclose(fp);
          }
        }
      }
      programEnd(status);
      return status;
    }

    STATUS chipVersion(VERSION& version)
    {
      STATUS status = protocol(_address, OTA_UPDATE_CHIPVER);
      version  = (VERSION)(status == STATUS_OK ? read32(_packet.data + 2) : 0);
      _otaboot = _packet.data[6];
      return status;
    }

    static const char *errstr(STATUS status)
    {
      switch (status)
      {
        case STATUS_OK:             return "Success";
        case STATUS_ARG_ERROR:      return "invalid argument";
        case STATUS_BLD_ERROR:      return "required bootloader";
        case STATUS_CRC_ERROR:      return "CRC error";
        case STATUS_SDK_ERROR:      return "SDK error";
        case STATUS_INVALID_DEVICE: return "not supported device";
        case STATUS_INVALID_FILE:   return "other device binaries (.bin)";
        case STATUS_OPEN_ERROR:     return "file open error";
        case STATUS_READ_ERROR:     return "file read error";
        case STATUS_TX_ERROR:       return "transmit error";
        case STATUS_RX_ERROR:       return "receive failed";
      }
      return "unknown error";
    }

  protected:

    virtual void programStart(const char *chipname, const char *filename, bool program) {}
    virtual void programProgress(size_t size, size_t processed) {}
    virtual void programEnd(STATUS status) {}

    STATUS command(uint16_t addr, uint8_t cmd, const uint8_t *buf = 0, uint8_t len = 0)
    {
      Bridge::packet_t packet;
      if (cmd)
      {
        packet.length  = len + offsetof(Bridge::packet_t, data[1]);
        packet.type    = 0;
        packet.addrH   = addr >> 8;
        packet.addrL   = addr >> 0;
        packet.data[0] = cmd;
        memcpy(packet.data + 1, buf, len);
      }
      else
      {
        packet.length = offsetof(Bridge::packet_t, data);
        packet.type   = Bridge::TYPE_RETRANSMIT;
        packet.addrH  = addr >> 8;
        packet.addrL  = addr >> 0;
      }
      *((uint8_t *)&packet + packet.length) = Bridge::csum((uint8_t *)&packet, packet.length);
      size_t cnt = packet.length + sizeof(packet.length);
#if DEBUG_TX_PACKET_DUMP
      printf("TX: ");
      dump((uint8_t *)&packet, cnt);
#endif
      return write((uint8_t *)&packet, cnt) == cnt ? STATUS_OK : STATUS_TX_ERROR;
    }

    STATUS response(void)
    {
      size_t n, cnt;
      for (cnt = _packet.length = 0; cnt <= _packet.length; cnt += n)
      {
        if ((n = read((uint8_t *)&_packet + cnt, _packet.length - cnt + sizeof(_packet.length))) == 0)
          break;
      }
#if DEBUG_RX_PACKET_DUMP
      printf("RX: ");
      dump((uint8_t *)&_packet, cnt);
#endif
      if (cnt != _packet.length + sizeof(_packet.length))
        return STATUS_RX_ERROR;
      return Bridge::csum((uint8_t *)&_packet, cnt) ? STATUS_CRC_ERROR : STATUS_OK;
    }

    const Bridge::packet_t *getPacket(void)
    {
      return &_packet;
    }

  private:

    bool fileVersion(VERSION& version, FILE *fp)
    {
      uint8_t data[4];
      fseek(fp, 0, SEEK_SET);
      if (fread(data, 1, sizeof(data), fp) != sizeof(data))
        return false;
      version = (VERSION)read32(data);
      return true;
    }

    void savePath(char *save, const char *path)
    {
      char c, *dir = save;
      while (*path)
      {
        c = *save++ = *path++;
        if ((c == '/') || (c == '\\'))
          dir = save;
      }
      *dir = 0;
    }

    void write32(uint8_t *buf, uint32_t data)
    {
      buf[3] = (uint8_t)(data >>  0);
      buf[2] = (uint8_t)(data >>  8);
      buf[1] = (uint8_t)(data >> 16);
      buf[0] = (uint8_t)(data >> 24);
    }

    uint32_t read32(const uint8_t *buf)
    {
      return
        ((uint32_t)buf[0] << 24) |
        ((uint32_t)buf[1] << 16) |
        ((uint16_t)buf[2] <<  8) |
        ((uint8_t )buf[3] <<  0);
    }

    void dump(const uint8_t *buf, uint8_t len)
    {
      int i = 0;
      while (i < len)
      {
        printf("%02X ", *buf++);
        if ((++i % 16) == 0)
          printf("\n");
      }
      if ((i % 16) != 0)
        printf("\n");
    }

    STATUS protocol(uint16_t addr, uint8_t cmd, const uint8_t *buf = 0, uint8_t len = 0)
    {
      STATUS status;
      for (int i = 0; i < OTA_RETRY_COUNT; ++i)
      {
        if ((status = command(addr, cmd, buf, len)) != STATUS_OK)
          return status;
        if ((status = response()) == STATUS_OK)
          return (STATUS)_packet.data[1];
        cmd = 0;
        len = 0;
        Sleep::sleep(100);
      }
      return status;
    }

    STATUS update(FILE *fp, size_t size, uint16_t addr, bool bldupd = false)
    {
      STATUS status;
      uint8_t len, buf[96];

      buf[0] = bldupd ? 0xBF : 0x00;
      if (!_otaboot && !bldupd)
        status = STATUS_BLD_ERROR;
      else if ((status = protocol(addr, OTA_UPDATE_BLDMODE, buf, sizeof(buf[0]))) == STATUS_OK)
      {
        if (!bldupd)
        {
          Sleep::sleep(500);
          addr = ADDR_OTA_BOOT;
        }
        if ((status = protocol(addr, OTA_UPDATE_START)) == STATUS_OK)
        {
          size_t progress = 0;
          size_t prog = 0;
          CRC32::init();
          size = (size + 15) & ~0x0F; // 16-byte boundary
          while (1)
          {
            memset(buf, 0xFF, sizeof(buf));
            if ((len = (size_t)fread(buf, 1, sizeof(buf), fp)) == 0)
              break;
            len = (len + 15) & ~0x0F; // 16-byte boundary
            if ((status = protocol(addr, OTA_UPDATE_PROGRAM, buf, len)) != STATUS_OK)
              return status;
            CRC32::update(buf, len);
            progress += len;
            if (prog != (progress >> 10))
            {
              prog = (progress >> 10);
              programProgress(size, progress);
            }
          }
          programProgress(size, progress);
        }
        write32(buf, CRC32::result());
        status = protocol(addr, OTA_UPDATE_END, buf, 4);
      }
      return status;
    }

    uint16_t _address;
    bool     _otaboot;
    Bridge::packet_t _packet;
};
