/*
 main.cpp - OTA Update Utility for NXP JN516x Series

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include "jn516x_ota.h"

class Device : public JN516x_OTA
{
  public:

    Device(void) {}
    virtual ~Device(void) {}

  protected:

    void programStart(const char *chipname, const char *filename, bool program) override
    {
      if (program)
        printf("Write Device [%s] <- \"%s\"\n", chipname, filename);
      else
        printf("Read Device [%s] -> \"%s\"\n", chipname, filename);
    }

    void programProgress(size_t size, size_t processed) override
    {
#if SIZE_MAX == UINT32_MAX // equals __SIZEOF_SIZE_T__ == 4
      printf("Progress: %6u/%6u\r", processed, size);
#else
      printf("Progress: %6llu/%6llu\r", processed, size);
#endif
      fflush(stdout);
    }

    void programEnd(STATUS status) override
    {
      printf("\n");
      if (status != STATUS_OK)
        printf("Error: [0x%02X] ", status);
      printf("%s\n", errstr(status));
    }
};

static Device device;

void chipInfo(void)
{
  Device::VERSION version;
  Device::STATUS status = device.chipVersion(version);
  if (status == Device::STATUS_OK)
  {
    printf("Chip Name: %s", device.chipName(version));
    printf(", Flash: %u Kb, Ram: %u Kb, Type: %u\n",
        device.chipFlashSize(version) / 1024,
        device.chipRamSize(version) / 1024,
        device.chipType(version));
  }
  else
    printf("%s\n", device.errstr(status));
}

void portScan(void)
{
  int count = device.getDeviceCount();
  for (int i = 0; i < count; ++i)
  {
    char devcom[64], devser[64], devdes[64];
    if (device.getComPortNumber(i) >= 0)
    {
      device.getComPortName(i, devcom);
      device.getDeviceSerial(i, devser);
      device.getDeviceDescription(i, devdes);
      printf("Device-%d: Port: %s, Serial: %s, Description: %s\n", i, devcom, devser, devdes);
    }
    else
      printf("Device-%d: -- Open failure --\n", i);
  }
}

void usage(bool help)
{
  printf("OTA Update Utility for NXP JN516X, Version 1.0\n");
  printf("Copyright(c) 2022 Sasapea's Lab. All right reserved.\n");
  printf("\n");
  if (help)
  {
    printf("Usage: JN516X-OTA [options]\n");
    printf("\n");
    printf("  options:\n");
    printf("    -p portname ... FTD2XX Serial Number for Open\n");
    printf("    -s          ... FTD2XX Port Scan\n");
    printf("    -w filename ... write device (.bin)\n");
    printf("    -a address  ... device address (*0 - 65534)\n");
    printf("    -d          ... device information\n");
    printf("    -h          ... this help\n");
    printf("\n");
  }
}

int main(int argc, char *argv[])
{
  const char *portname = "";
  const char *firmware = "";
  unsigned long speed = 1000000;
  uint16_t address = 0;
  bool scan = false;
  bool info = false;
  bool help = false;

  for (int i = 1; i < argc; )
  {
    char *p = argv[i++];
    if (*p == '-')
    {
      switch (*++p)
      {
        case 'h':
          help = true;
          break;
        case 's':
          scan = true;
          break;
        case 'p':
          if (i < argc)
            portname = argv[i++];
          break;
        case 'w':
          if (i < argc)
            firmware = argv[i++];
          break;
        case 'a':
          if (i < argc)
            address = atoi(argv[i++]);
          break;
        case 'd':
          info = true;
          break;
        default:
          help = true;
          break;
      }
    }
    else
      help = true;
  }

  if (help)
    usage(true);
#ifndef __linux__
  else if (device.status() != FT_OK)
  {
    usage(false);
    printf("FTD2XX Library Not Found.");
  }
#endif
  else if (scan)
  {
    usage(false);
    portScan();
  }
  else if (device.open(portname, speed, address))
  {
    usage(false);
    if (info)
      chipInfo();
    if (*firmware)
      device.update(firmware);
    device.close();
  }
  else
  {
    usage(false);
    printf("Port Open Failed: \"%s\"\n", portname);
  }

  return 0;
}
